/* -*- C -*- 

   memP Heap Profiler ( http://memp.sourceforge.net/ )

   Please see COPYRIGHT AND LICENSE information at the end of this file.

   ----- 

   pointer_tracking.c -- track pointers to allocated memory

 */

#ifndef lint
static char *svnid =
  "$Id: pointer_tracking.c 369 2006-10-05 19:21:12Z chcham $";
#endif

#include "memPconfig.h"
#include "memPi.h"
#include "symbols.h"
#include <string.h>

/******************************************************************************/
/*                     POINTER HASH TABLE IMPLEMENTATION                      */
/******************************************************************************/


typedef struct _alloc_size_t
{
  void *ptr;
  size_t sz;
  callsite_stats_t *csp;
}
alloc_size_t;

extern void *(*memPi_malloc_ptr) (size_t);

unsigned int
memPi_alloc_ptr_hashkey (const void *p)
{
  alloc_size_t *pa = (alloc_size_t *) p;
  return 52271 ^ (unsigned int) pa->ptr;
}

int
memPi_alloc_ptr_comparator (const void *p1, const void *p2)
{
  alloc_size_t *pa1 = (alloc_size_t *) p1;
  alloc_size_t *pa2 = (alloc_size_t *) p2;

  memPi_msg_debug_level (MEMP_DEBUG_ALLOC,
                         "Comparing allocation pointers %p and %p\n",
                         pa1->ptr, pa2->ptr);

  if (pa1->ptr > pa2->ptr)
    return 1;

  if (pa1->ptr < pa2->ptr)
    return -1;

  return 0;
}

void
memPi_open_alloc_hash ()
{
  /* h_open will abort if unable to allocate hash table */
  memPi.alloc_ptr_hash =
    h_open (memPi.tableSize, memPi_alloc_ptr_hashkey,
            memPi_alloc_ptr_comparator);
}

void
memPi_insert_alloc_ptr (h_t * ht, void *retval, size_t msz,
                        alloc_stats_t * as, callsite_stats_t * csp)
{
  alloc_size_t *asp;
  alloc_size_t key;

  key.ptr = retval;
  if (h_search (ht, &key, (void **) &asp) != NULL)
    memPi_msg_warn ("Found existing hash entry for pointer %p\n", retval);
  else
  {
    asp = (alloc_size_t *) memPi_malloc_ptr (sizeof (alloc_size_t));
    asp->ptr = retval;
    asp->sz = msz;
    asp->csp = csp;

    if (h_insert (ht, asp) != 0)
      memPi_msg_warn
        ("Found hash table inconsistency for h_insert at memPi_insert_alloc_ptr\n");

    memPi_msg_debug_level (MEMP_DEBUG_ALLOC,
                           "Rank %4d : malloc added pointer %p size %lu\n",
                           memPi.rank, asp->ptr, asp->sz);
    as->alloc_cnt++;
    as->alloc_tot += msz;
    as->alloc_cur += msz;
    if (as->alloc_cur > as->alloc_max)
      as->alloc_max = as->alloc_cur;
  }
}


long long int
memPi_update_alloc_ptr (h_t * ht, void *inptr, void *newptr, size_t msz,
                        alloc_stats_t * as, callsite_stats_t * csp)
{
  alloc_size_t *asp;
  alloc_size_t key;
  size_t prev_size;
  long long int retval = 0;

  key.ptr = inptr;

  if (h_search (ht, &key, (void **) &asp) != NULL)
  {
    memPi_msg_debug_level (MEMP_DEBUG_ALLOC,
                           "update found alloc of size %ld at ptr %p\n",
                           asp->sz, asp->ptr);
    prev_size = asp->sz;
    asp->sz = msz;
    if (inptr != newptr)
    {
      h_delete (ht, &key, (void **) &asp);
      asp->ptr = newptr;
      h_insert (ht, asp);
    }

    memPi_msg_debug_level (MEMP_DEBUG_ALLOC,
                           "updates pointer %p from %ld to %ld\n", asp->ptr,
                           prev_size, msz);

    as->alloc_cnt++;
    as->alloc_tot += msz - prev_size;
    as->alloc_cur += msz - prev_size;
    asp->csp->csCurAlloc += (long long int) msz - (long long int) prev_size;
    if (as->alloc_cur > as->alloc_max)
      as->alloc_max = as->alloc_cur;

    retval = (long long int) msz - (long long int) prev_size;
  }
  else
  {
    memPi_insert_alloc_ptr (memPi.alloc_ptr_hash, newptr, msz, as, csp);
    memPi_msg_warn ("Failed to find ptr %p to update!\n");
  }
  return retval;
}

size_t
memPi_delete_alloc_ptr (h_t * ht, void *inptr, alloc_stats_t * as)
{
  alloc_size_t *asp;
  alloc_size_t key;
  size_t retval = 0;

  if (inptr == NULL)
    return 0;

  key.ptr = inptr;

  if (h_search (ht, &key, (void **) &asp) != NULL)
  {
    memPi_msg_debug_level (MEMP_DEBUG_ALLOC,
                           "delete found alloc of size %ld at ptr %p\n",
                           asp->sz, asp->ptr);

    retval = asp->sz;

    as->free_cnt++;
    as->free_tot += asp->sz;
    as->alloc_cur -= asp->sz;
    if (asp->csp != NULL)
      asp->csp->csCurAlloc -= asp->sz;
    else if (memPi.traceAllStacks == 1)
      memPi_msg_warn ("Failed to find callsite for allocation!\n");

    h_delete (ht, &key, (void **) &asp);
    memPi_free (asp);
  }
  else
  {
    /*  Assume that a pointer that is not found has been allocated prior to MPI_Init  
       and attempt to determine the size.

       TODO: what if this results in SEGFAULT?
     */
    size_t ps = *(((size_t *) inptr) - 1);

    memPi_msg_debug ("Delete failed to find ptr %p of size %u to delete!\n",
                     inptr, ps);

    /* Check for underflow, which would indicate erroneous unknown value
       or incorrect malloc size.
     */
    if (ps < memPi.unknownHeapUse)
    {
      memPi.unknownHeapUse -= ps;
      memPi_msg_debug ("Reduced unknownHeapUse by %u to %u\n", ps,
                       memPi.unknownHeapUse);
    }

    retval = 0;
  }

  return retval;
}


/* 

<license>

Copyright (c) 2010, Lawrence Livermore National Security, LLC.
Produced at the Lawrence Livermore National Laboratory
Written by Christopher Chambreau <chcham@llnl.gov>.
CODE-OCEC-10-010
All rights reserved.

This file is part of memP. For details, see http://memp.sourceforge.net .
Please also read this link – Additional BSD Notice.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  o  Redistributions of source code must retain the above copyright notice, 
     this list of conditions and the disclaimer below.

  o  Redistributions in binary form must reproduce the above copyright notice, 
     this list of conditions and the disclaimer (as noted below) in the 
     documentation and/or other materials provided with the distribution.

  o  Neither the name of the LLNS/LLNL nor the names of its contributors may be
     used to endorse or promote products derived from this software without 
     specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL LAWRENCE LIVERMORE NATIONAL
SECURITY, LLC, THE U.S. DEPARTMENT OF ENERGY OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
THE POSSIBILITY OF SUCH DAMAGE.


Additional BSD Notice

1. This notice is required to be provided under our contract with the U.S.
Department of Energy (DOE). This work was produced at Lawrence Livermore
National Laboratory under Contract No. DE-AC52-07NA27344 with the DOE.

2. Neither the United States Government nor Lawrence Livermore National
Security, LLC nor any of their employees, makes any warranty, express or
implied, or assumes any liability or responsibility for the accuracy,
completeness, or usefulness of any information, apparatus, product, or process
disclosed, or represents that its use would not infringe privately-owned rights.

3. Also, reference herein to any specific commercial products, process, or
services by trade name, trademark, manufacturer or otherwise does not
necessarily constitute or imply its endorsement, recommendation, or favoring by
the United States Government or Lawrence Livermore National Security, LLC.
The views and opinions of authors expressed herein do not necessarily state or
reflect those of the United States Government or Lawrence Livermore National
Security, LLC, and shall not be used for advertising or product endorsement
purposes.

</license>

*/


/* eof */
