/* -*- C -*- 

   memP Heap Profiler ( http://memp.sourceforge.net/ )

   Please see COPYRIGHT AND LICENSE information at the end of this file.

   ----- 

   record_stack.c -- Implementations of memPi_RecordTraceBack 
                       stack tracing function.

 */

#ifndef lint
static char *svnid = "$Id: record_stack.c 403 2007-04-26 15:15:06Z chcham $";
#endif

#include <setjmp.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>


#include "memPi.h"

#ifdef HAVE_LIBUNWIND
#define UNW_LOCAL_ONLY
#include <libunwind.h>
#endif

int
memPi_get_stack_trace (void *pc_array[MEMP_CALLSITE_STACK_DEPTH_MAX])
{
  int frame_count = 0;
  const unsigned int stack_offset = memPi.skipFrames;

  frame_count =
    memP_record_traceback (pc_array, MEMP_CALLSITE_STACK_DEPTH_MAX);

#if 1
  //  Remove PC for wrapper function
  memmove (pc_array, &pc_array[stack_offset],
           sizeof (void *) * (frame_count - stack_offset));
  pc_array[frame_count - stack_offset] = NULL;
#endif

  //memPi_print_stack(pc_array);

  return frame_count;
}



#ifdef HAVE_LIBUNWIND

int
memPi_RecordTraceBack (jmp_buf jb, void *pc_array[], int max_back)
{
  int i, valid_cursor, frame_count = 0, retval;
  unw_context_t uc;
  unw_cursor_t cursor;
  unw_word_t pc;

  if ((retval = unw_getcontext (&uc)) != 0)
  {
    memPi_msg_debug_level (MEMP_DEBUG_STACK, "Failed unw_getcontext: %s\n",
                           unw_strerror (retval));
    return frame_count;
  }

  if ((retval = unw_init_local (&cursor, &uc)) != UNW_ESUCCESS)
  {
    memPi_msg_debug_level
      (MEMP_DEBUG_STACK,
       "Failed to initialize libunwind cursor with unw_init_local: %s\n",
       unw_strerror (retval));
  }
  else
  {
    /*  Unwind this function before recording stack trace  */
    if ((retval = unw_step (&cursor)) < 1)
    {
      if (retval < 0)
        memPi_msg_debug_level (MEMP_DEBUG_STACK,
                               "First unw_step failed: %s.\n",
                               unw_strerror (retval));

      return frame_count;
    }

    for (i = 0, valid_cursor = 1; i < max_back; i++)
    {
      if (valid_cursor && (retval = unw_step (&cursor)) > 0)
      {
        int reg_retval;

        frame_count++;
        if ((reg_retval =
             unw_get_reg (&cursor, UNW_REG_IP, &pc)) != UNW_ESUCCESS)
        {
          pc_array[i] = NULL;
          memPi_msg_debug_level (MEMP_DEBUG_STACK, "unw_get_reg failed: %s\n",
                                 unw_strerror (reg_retval));
        }
        else
        {
          pc_array[i] = (void *) ((char *) pc - 1);
        }
      }
      else
      {
        pc_array[i] = NULL;
        valid_cursor = 0;
        if (retval < 0)
          memPi_msg_debug_level (MEMP_DEBUG_STACK, "unw_step failed: %s.\n",
                                 unw_strerror (retval));
      }
    }
  }

  return frame_count;
}

#else /* not LIBUNWIND */

#ifdef OSF1

int
memPi_RecordTraceBack (jmp_buf jb, void *pc_array[], int max_back)
{
  int i, parent_frame_start, frame_count = 0;
  void *pc;
  PCONTEXT context;

  context = (PCONTEXT) jb;

  parent_frame_start = 2;

  for (i = 1; i < parent_frame_start; i++)
  {
    pc = (void *) context->sc_pc;
    if (pc != NULL)
      unwind (context, 0);
  }

  for (i = 0; i < max_back; i++)
  {
    if (pc != NULL)
    {
      unwind (context, 0);

      /* record this frame's pc and calculate the previous instruction  */
      pc_array[i] = (void *) context->sc_pc - (void *) 0x4;
      pc = (void *) context->sc_pc;
      frame_count++;
    }
    else
    {
      pc_array[i] = NULL;
    }
  }

  return frame_count;
}

#elif defined(mips) && defined(__GNUC__)
  /* on the MIPS we use gcc's predefined macro to get a single level
   * of stack backtrace
   */
int
memPi_RecordTraceBack (jmp_buf jb, void *pc_array[], int max_back)
{
  memPi_msg_debug_level
    (MEMP_DEBUG_STACK,
     "Using saved address on this MIPS arch (max traceback=1)\n");
  if (max_back == 0)
    return 0;

  /* For MIPS under GNUC, we can only handle a traceback upto one level */
  if (max_back > 1)
  {
    memPi_msg_warn
      ("We only support a single level of stack backtrace on MIPS currently\n"
       "Using the a traceback of 1, instead\n");
  }
  pc_array[0] = (void *) ((char *) saved_ret_addr - 1);
  pc_array[1] = '\0';
  return 1;
}

#elif defined(HAVE_STACKWALKER)
#include <vector>
#include "walker.h"
#include "frame.h"
using namespace Dyninst;
using namespace
  Dyninst::Stackwalker;

extern
  "C" int
memPi_RecordTraceBack (jmp_buf jb, void *pc_array[], int max_back)
{
  std::vector < Frame > stackwalk;
  Walker *
    walker = Walker::newWalker ();
  int
    frame_count;

  if (walker->walkStack (stackwalk) == false)
    return 0;

  frame_count = stackwalk.size ();

  for (unsigned i = 0; i < frame_count && i < (max_back - 1); i++)
  {
    fprintf (stderr, "Got RA for frame %d of %x\n", i,
             (unsigned int) stackwalk[i].getRA ());
    pc_array[i] = (void *) stackwalk[i].getRA ();
  }

  return frame_count;
}

#else /* not OSF1 and not MIPS+GCC */

int
memPi_RecordTraceBack (jmp_buf jb, void *pc_array[], int max_back)
{
  int i, frame_count = 0;
  void *fp, *lastfp;
  void *pc = (void *) 1;

  /*  
     For standard memP, the current stack looks like:

     memPi_RecordTraceBack 
     memP_record_traceback <- setjmp called
     get_stack_trace
     malloc wrapper
     Application function
     ...

     with the jmpbuf set in memP_record_traceback
   */


#if !defined(UNICOS_mp)
  /*  This gets the FP for the Application function from
     the setjmp jmp_buf.  */
  fp = ParentFP (jb);
#else
  /*  For UNICOS, we first get the frame pointer for what would
     be the application function and step through an additional frame.
   */
  fp = NextFP (GetFP ());
  fp = NextFP (fp);

#endif /* defined(UNICOS_mp) */

  for (i = 0; i < max_back; i++)
  {
    if (fp != NULL && pc != NULL)
    {
      /* record this frame's pc */
      pc = FramePC (fp);
      if (pc != NULL)
      {
        frame_count++;
        /*  Get previous instruction  */
        pc_array[i] = (void *) ((char *) pc - 1);
      }
      else
        pc_array[i] = NULL;

      /* update frame ptr */
      lastfp = fp;
      fp = NextFP (fp);
    }
    else
    {
      /* pad the array w/ nulls, if necessary */
      pc_array[i] = NULL;
    }
  }

  return frame_count;
}
#endif /* OSF1 */
#endif /* libunwind */


#define USE_BACKTRACE
#if !defined(USE_BACKTRACE)
#ifdef HAVE_STACKWALKER
extern "C"
#endif
  int
memP_record_traceback (void *pc_array[], int max_stack)
{
  jmp_buf jb;
  int retval;

  setjmp (jb);

  retval = memPi_RecordTraceBack (jb, pc_array, max_stack);

  return retval;
}
#else

#include <execinfo.h>

int
memP_record_traceback (void *pc_array[], int max_stack)
{
  return backtrace (pc_array, max_stack);
}

#endif


/* 

<license>

Copyright (c) 2010, Lawrence Livermore National Security, LLC.
Produced at the Lawrence Livermore National Laboratory
Written by Christopher Chambreau <chcham@llnl.gov>.
CODE-OCEC-10-010
All rights reserved.

This file is part of memP. For details, see http://memp.sourceforge.net .
Please also read this link – Additional BSD Notice.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  o  Redistributions of source code must retain the above copyright notice, 
     this list of conditions and the disclaimer below.

  o  Redistributions in binary form must reproduce the above copyright notice, 
     this list of conditions and the disclaimer (as noted below) in the 
     documentation and/or other materials provided with the distribution.

  o  Neither the name of the LLNS/LLNL nor the names of its contributors may be
     used to endorse or promote products derived from this software without 
     specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL LAWRENCE LIVERMORE NATIONAL
SECURITY, LLC, THE U.S. DEPARTMENT OF ENERGY OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
THE POSSIBILITY OF SUCH DAMAGE.


Additional BSD Notice

1. This notice is required to be provided under our contract with the U.S.
Department of Energy (DOE). This work was produced at Lawrence Livermore
National Laboratory under Contract No. DE-AC52-07NA27344 with the DOE.

2. Neither the United States Government nor Lawrence Livermore National
Security, LLC nor any of their employees, makes any warranty, express or
implied, or assumes any liability or responsibility for the accuracy,
completeness, or usefulness of any information, apparatus, product, or process
disclosed, or represents that its use would not infringe privately-owned rights.

3. Also, reference herein to any specific commercial products, process, or
services by trade name, trademark, manufacturer or otherwise does not
necessarily constitute or imply its endorsement, recommendation, or favoring by
the United States Government or Lawrence Livermore National Security, LLC.
The views and opinions of authors expressed herein do not necessarily state or
reflect those of the United States Government or Lawrence Livermore National
Security, LLC, and shall not be used for advertising or product endorsement
purposes.

</license>

*/


/* eof */
