/* -*- C -*- 

   memP Heap Profiler ( http://memp.sourceforge.net/ )

   Please see COPYRIGHT AND LICENSE information at the end of this file.

   ----- 

   statistics.c -- track allocation information

 */

#ifndef lint
static char *svnid = "$Id: statistics.c 383 2006-12-08 18:30:15Z chcham $";
#endif

#include <string.h>
#include <strings.h>
#include <float.h>
#include <limits.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/time.h>
#include <sys/resource.h>
#include "memPi.h"

long long task_alloc_in_use;
long long task_report_count;

callsite_stats_t *
memPi_update_callsite_stats (unsigned op, int rank, void **pc,
                             double dur, double allocSize, double freeSize)
{
  int i;
  callsite_stats_t *csp = NULL;
  callsite_stats_t key;

  if (!memPi.enabled)
    return NULL;

  assert (memPi.task_callsite_stats != NULL);
  assert (dur >= 0);

  key.op = op;
  key.rank = rank;
  key.cookie = MEMP_CALLSITE_STATS_COOKIE;
  for (i = 0; i < MEMP_CALLSITE_STACK_DEPTH; i++)
  {
    key.pc[i] = pc[i];
  }

  if (NULL == h_search (memPi.task_callsite_stats, &key, (void **) &csp))
  {
    /* create and insert */
    csp = (callsite_stats_t *) memPi_malloc (sizeof (callsite_stats_t));
    bzero (csp, sizeof (callsite_stats_t));
    csp->op = op;
    csp->rank = rank;
    for (i = 0; i < MEMP_CALLSITE_STACK_DEPTH; i++)
    {
      csp->pc[i] = pc[i];
    }
    csp->cookie = MEMP_CALLSITE_STATS_COOKIE;
    csp->minAlloc = DBL_MAX;
    csp->minFree = DBL_MAX;
    csp->csLWM = (double) (1 << 30);
    if (h_insert (memPi.task_callsite_stats, csp) == 1)
      memPi_msg_warn
        ("Duplicate insertion found in memPi.task_callsite_stats\n");

  }
  /* ASSUME: csp cannot be deleted from list */

  //  Standard allocation metrics
  csp->count++;
  csp->cumulativeTime += dur;

  csp->cumulativeAlloc += allocSize;
  csp->cumulativeFree += freeSize;

  //  Individual allocation statistics  
  csp->maxAlloc = max (csp->maxAlloc, allocSize);
  csp->minAlloc = min (csp->minAlloc, allocSize);

  csp->maxFree = max (csp->maxFree, freeSize);
  csp->minFree = min (csp->minFree, freeSize);

  /* HWM allocation parameters to track:
   * 
   * Task HWM
   * Alloc size at Task HWM
   * Call site HWM
   * Task allocation count at task HWM
   * Call site allocation count at task HWM
   * Run time timestamp at task HWM
   * Task LWM at call site
   */

  //  Update current amount in use by this call site
  csp->csCurAlloc += allocSize;

  /* In the case of realloc where a free and an allocation occurred, do not
   * subtract the free size from the current allocation
   */

  if (allocSize <= 0)
    csp->csCurAlloc -= freeSize;

//  Do the task-wide updating independently of call site activity
#if 0
  //  Update task and call site HWMs
  if ((double) (task_alloc_in_use + memPi.unknownHeapUse) > memPi.taskHWM)
  {
    memPi.taskHWM = (double) task_alloc_in_use + memPi.unknownHeapUse;
    memPi.taskHWMunknown = (double) memPi.unknownHeapUse;
    //fprintf(stderr, "Setting rank %d hwm to %f\n", memPi.rank, memPi.taskHWM);
    memPi.taskHWMsize = allocSize;
    memPi.taskHWMtaskCount = (double) memPi.task_alloc_cnt;
    memPi.taskHWMcsCount = (double) csp->count;
    memPi_GETTIME (&memPi.taskHWMts);

    memPi.taskHWMop = op;
    for (i = 0; i < MEMP_CALLSITE_STACK_DEPTH; i++)
    {
      memPi.taskHWMpcs[i] = csp->pc[i];
    }
    /* If current allocation meets individual task reporting
     * criteria, generate a task-specific memory in-use report.
     */
    if (memPi.enabled == 1
        && task_report_count < memPi.taskHWMreportMaxCount
        && (double) task_alloc_in_use > memPi.taskHWMreportThreshold
        && (memPi.rank == memPi.taskHWMrank || memPi.taskHWMrank == -1))
    {
      memPi_profile_print_task_xml_verbose ();
      if (memPi.abortAtThreshold == 1)
        memPi_abort
          ("High Water Mark Threshold of %.0f bytes exceeded.\n",
           memPi.taskHWMreportThreshold);
      task_report_count++;
    }
  }
#endif
  csp->csHWM = max (csp->csHWM, csp->csCurAlloc);
  csp->csLWM =
    min (csp->csLWM, (double) task_alloc_in_use + memPi.unknownHeapUse);

#if 0
  memPi_msg_debug ("memPi.messageCountThreshold is %d\n",
                   memPi.messageCountThreshold);
  memPi_msg_debug ("sendSize is %f\n", allocSize);
  memPi_msg_debug ("csp->arbitraryMessageCount is %lld\n",
                   csp->arbitraryMessageCount);
#endif

  return csp;
}


#ifdef HAVE_BLUE_GENE_KERNEL_INTERFACE
#include <stdio.h>
#include <spi/kernel_interface.h>
#endif

void
memPi_update_task_stats (size_t allocSize, size_t freeSize,
                         callsite_stats_t * csp, unsigned int op)
{
  /***********************************************************
    Update global allocation statistics.
   **********************************************************/
  size_t currstack, currheaptot, currheapavail;
  MEMP_TRACE ("Enter memPi_update_task_stats");

#ifdef HAVE_BLUE_GENE_KERNEL_INTERFACE

  Kernel_GetMemorySize (KERNEL_MEMSIZE_STACK, &currstack);
  //Kernel_GetMemorySize(KERNEL_MEMSIZE_HEAP, &currheaptot);

  memPi_msg_debug
    ("Kernel says currstack = %u, currheaptot = %u, currheapavail = %u\n",
     currstack, currheaptot, currheapavail);
#else

  MEMP_TRACE ("In memPi_update_task_stats before checking for topOfStack");
  if ((void *) &currstack > memPi.topOfStack)
    memPi.topOfStack = (void *) &currstack;

  memPi_msg_debug ("Calculating stack size as %p - %p = %u\n",
                   memPi.topOfStack, &currstack,
                   (size_t) (memPi.topOfStack - (void *) &currstack));
  currstack = memPi.topOfStack - (void *) &currstack;
  MEMP_TRACE ("In memPi_update_task_stats after checking for topOfStack");

#endif

#ifdef DO_GET_AVAIL
  MEMP_TRACE ("In memPi_update_task_stats before memPi_get_heap_avail");
  currheapavail = memPi_get_heap_avail (0);
  MEMP_TRACE ("In memPi_update_task_stats after memPi_get_heap_avail");
#endif

  /* Track tot current heap use */
  task_alloc_in_use += allocSize;

  /* Only subtract freeSize from current heap use if the result would be > 0,
     otherwise unsigned overflow results in incorrect HWM.
   */
  if (task_alloc_in_use - freeSize < 0)
  {
    memPi_msg_warn
      ("free would cause task heap use < 0, setting heap use to 0");
    task_alloc_in_use = 0;
  }
  else
    task_alloc_in_use -= freeSize;

  /* Track total number of allocations */
  if (allocSize > 0)
    memPi.task_alloc_cnt++;

#if 0
  if (memPi.enabled == 0)
    return;
#endif

  /***********************************************************
    Update HWM and call site statistics
   **********************************************************/
  MEMP_TRACE ("In memPi_update_task_stats before updating stats");
  if ((double) (task_alloc_in_use + memPi.unknownHeapUse) > memPi.taskHWM)
  {
    /*fprintf(stdout, "in_use: %llu, unknown: %d\n", task_alloc_in_use, memPi.unknownHeapUse); */
    memPi.taskHWM = (double) task_alloc_in_use + memPi.unknownHeapUse;
    memPi.taskHWMunknown = (double) memPi.unknownHeapUse;

    memPi.taskHWMsize = allocSize;
    memPi.taskHWMtaskCount = (double) memPi.task_alloc_cnt;
    memPi.taskHWMop = op;
    memPi.taskHWMheapAvail = (double) currheapavail;
    memPi.taskHWMstackUse = currstack;
    memPi_GETTIME (&memPi.taskHWMts);

    if (csp != NULL)
    {
      int i;

      memPi.taskHWMcsCount = (double) csp->count;
      for (i = 0; i < MEMP_CALLSITE_STACK_DEPTH; i++)
      {
        memPi.taskHWMpcs[i] = csp->pc[i];
      }
    }
    else if (allocSize > 0 && memPi.traceAllStacks == 1 && memPi.inWrappedFunc == 1)
      memPi_msg_warn ("Received NULL call site pointer to update HWM!\n");

    else if (memPi.traceAllStacks == 0 && memPi.inWrappedFunc == 1 && memPi.afterInit == 1)     /* Just do stack trace for new HWM */
      memPi_get_stack_trace (memPi.taskHWMpcs);

    memPi_msg_debug ("New HWM for alloc of %u.  mallinfo:\n",
                     (size_t) allocSize);

    if (memPi_debug > 0 && memPi.afterInit == 1)
      print_mallinfo ();
  }
  MEMP_TRACE ("In memPi_update_task_stats after updating stats");


  MEMP_TRACE ("In memPi_update_task_stats before report test");
  /* If current allocation meets individual task reporting
   * criteria, generate a task-specific memory in-use report.
   */
  if (memPi.enabled == 1 &&
      memPi.afterInit == 1 &&
      memPi.inWrappedFunc == 1 &&
      task_report_count < memPi.taskHWMreportMaxCount
      && (double) task_alloc_in_use > memPi.taskHWMreportThreshold
      && (memPi.rank == memPi.taskHWMrank || memPi.taskHWMrank == -1))
  {
    memPi_profile_print_task_xml_verbose ();
    if (memPi.abortAtThreshold == 1)
      memPi_abort
        ("High Water Mark Threshold of %.0f bytes exceeded.\n",
         memPi.taskHWMreportThreshold);
    task_report_count++;
  }
  MEMP_TRACE ("In memPi_update_task_stats after report test");
  MEMP_TRACE ("Exit memPi_update_task_stats");
}


/* 

<license>

Copyright (c) 2010, Lawrence Livermore National Security, LLC.
Produced at the Lawrence Livermore National Laboratory
Written by Christopher Chambreau <chcham@llnl.gov>.
CODE-OCEC-10-010
All rights reserved.

This file is part of memP. For details, see http://memp.sourceforge.net .
Please also read this link – Additional BSD Notice.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  o  Redistributions of source code must retain the above copyright notice, 
     this list of conditions and the disclaimer below.

  o  Redistributions in binary form must reproduce the above copyright notice, 
     this list of conditions and the disclaimer (as noted below) in the 
     documentation and/or other materials provided with the distribution.

  o  Neither the name of the LLNS/LLNL nor the names of its contributors may be
     used to endorse or promote products derived from this software without 
     specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL LAWRENCE LIVERMORE NATIONAL
SECURITY, LLC, THE U.S. DEPARTMENT OF ENERGY OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
THE POSSIBILITY OF SUCH DAMAGE.


Additional BSD Notice

1. This notice is required to be provided under our contract with the U.S.
Department of Energy (DOE). This work was produced at Lawrence Livermore
National Laboratory under Contract No. DE-AC52-07NA27344 with the DOE.

2. Neither the United States Government nor Lawrence Livermore National
Security, LLC nor any of their employees, makes any warranty, express or
implied, or assumes any liability or responsibility for the accuracy,
completeness, or usefulness of any information, apparatus, product, or process
disclosed, or represents that its use would not infringe privately-owned rights.

3. Also, reference herein to any specific commercial products, process, or
services by trade name, trademark, manufacturer or otherwise does not
necessarily constitute or imply its endorsement, recommendation, or favoring by
the United States Government or Lawrence Livermore National Security, LLC.
The views and opinions of authors expressed herein do not necessarily state or
reflect those of the United States Government or Lawrence Livermore National
Security, LLC, and shall not be used for advertising or product endorsement
purposes.

</license>

*/


/* eof */
